//----------------------------------------------------------------------
// File:			bd_tree.h
// Programmer:		David Mount, Andrei Sobolevskii, Aleksandr Andrievskii
// Description:		Declarations for standard bd-tree routines
// Last modified:	06/22/08
//----------------------------------------------------------------------
// Copyright 2008 Andrei Sobolevskii and Aleksandr Andrievskii
//
// Copyright (c) 1997-2005 University of Maryland and Sunil Arya and
// David Mount.  All Rights Reserved.
// 
// This file is part of the WANN (Weighted Approximate Nearest 
// Neighbors) library.
//
// WANN is free software: you can redistribute it and/or modify it
// under the terms of the GNU Lesser General Public License as
// published by the Free Software Foundation, either version 3 of the
// License, or (at your option) any later version.
//
// WANN is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
// or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General
// Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public
// License along with WANN.  If not, see <http://www.gnu.org/licenses/>.
//----------------------------------------------------------------------
// Acknowledgment:
//
// The development of the WANN library has been supported by ANR
// (Agence Nationale de la Recherche, France) under contract 
// BLAN07-2_183172 OTARIE and by RFBR (Russian Foundation for Basic 
// Research) under contract RFBR/CNRS 07-01-92217.
//----------------------------------------------------------------------
// History:
//
// This file has been originally released under GNU LGPL as part of the ANN
// (Approximate Nearest Neighbors) library of David Mount and Sunil
// Arya.  It has been modified to become part of the WANN library.
//
// WANN history:
//	Revision 0.2 06/22/08
//		New args to ann_search, ann_pri_search, ann_FR_search
//	Revision 0.1 06/03/08
//		Modified copyright information.
//		Changes comments.
//		Added pidx to ANNbd_shrink.
//		Added setWeight method.
//
// ANN history:
//	Revision 1.0  04/01/05
//		Changed IN, OUT to ANN_IN, ANN_OUT
//	Revision 0.1  03/04/98
//		Initial release
//----------------------------------------------------------------------

#ifndef ANN_bd_tree_H
#define ANN_bd_tree_H

#include <wann/ANNx.h>					// all ANN includes
#include "kd_tree.h"					// kd-tree includes

//----------------------------------------------------------------------
//	bd-tree shrinking node.
//		The main addition in the bd-tree is the shrinking node, which
//		is declared here.
//
//		Shrinking nodes are defined by list of orthogonal halfspaces.
//		These halfspaces define a (possibly unbounded) orthogonal
//		rectangle.  There are two children, in and out.  Points that
//		lie within this rectangle are stored in the in-child, and the
//		other points are stored in the out-child.
//
//		We use a list of orthogonal halfspaces rather than an
//		orthogonal rectangle object because typically the number of
//		sides of the shrinking box will be much smaller than the
//		worst case bound of 2*dim.
//
//		BEWARE: Note that constructor just copies the pointer to the
//		bounding array, but the destructor deallocates it.  This is
//		rather poor practice, but happens to be convenient.  The list
//		is allocated in the bd-tree building procedure rbd_tree() just
//		prior to construction, and is used for no other purposes.
//
//		WARNING: In the near neighbor searching code it is assumed that
//		the list of bounding halfspaces is irredundant, meaning that there
//		are no two distinct halfspaces in the list with the same outward
//		pointing normals.
//----------------------------------------------------------------------

class ANNbd_shrink : public ANNkd_node	// splitting node of a kd-tree
{
	int					n_bnds;			// number of bounding halfspaces
	ANNorthHSArray		bnds;			// list of bounding halfspaces
	ANNkd_ptr			child[2];		// in and out children
public:
	ANNbd_shrink(						// constructor
		int				nb,				// number of bounding halfspaces
		ANNorthHSArray	bds,			// list of bounding halfspaces
#ifndef ANN_NO_WEIGHTS
		ANNkd_ptr ic=NULL, ANNkd_ptr oc=NULL,	// children
		ANNidxArray pidx=NULL)					// union of buckets
#else
		ANNkd_ptr ic=NULL, ANNkd_ptr oc=NULL)	// children
#endif
		{
			n_bnds			= nb;				// cutting dimension
			bnds			= bds;				// assign bounds
			child[ANN_IN]	= ic;				// set children
			child[ANN_OUT]	= oc;
#ifndef ANN_NO_WEIGHTS
			bkt = pidx;
#endif
		}

	~ANNbd_shrink()						// destructor
		{
			if (child[ANN_IN]!= NULL && child[ANN_IN]!=  KD_TRIVIAL) 
				delete child[ANN_IN];
			if (child[ANN_OUT]!= NULL&& child[ANN_OUT]!= KD_TRIVIAL) 
				delete child[ANN_OUT];
			if (bnds != NULL)
				delete [] bnds;			// delete bounds
		}

	virtual void getStats(						// get tree statistics
				int dim,						// dimension of space
				ANNkdStats &st,					// statistics
				ANNorthRect &bnd_box);			// bounding box
	virtual void print(int level, ostream &out);// print node
	virtual void dump(ostream &out);			// dump node

	virtual void ann_search(ANNdist box_dist	// standard search
#ifdef WANN_TORUS
							, ANNpoint q 
#endif    
    );			

	virtual void ann_pri_search(ANNdist box_dist // priority search
#ifdef WANN_TORUS
       , ANNpoint q 
#endif    
    );		

	virtual void ann_FR_search(ANNdist box_dist	// fixed-radius search
#ifdef WANN_TORUS
							   , ANNpoint q 
#endif    
    ); 		

#ifndef ANN_NO_WEIGHTS
	virtual void set_weight						// set weight of point
		(ANNidx* idx_ptr, ANNdist new_weight, ANNdistArray weights);
#endif
};

#endif
