//----------------------------------------------------------------------
// File:			brute.cxx
// Programmer:		Sunil Arya and David Mount, 
//					Andrei Sobolevskii and Aleksandr Andrievskii
// Description:		Brute-force nearest neighbors
// Last modified:	06/22/08
//----------------------------------------------------------------------
// Copyright 2008 Andrei Sobolevskii and Aleksandr Andrievskii
//
// Copyright (c) 1997-2005 University of Maryland and Sunil Arya and
// David Mount.  All Rights Reserved.
// 
// This file is part of the WANN (Weighted Approximate Nearest 
// Neighbors) library.
//
// WANN is free software: you can redistribute it and/or modify it
// under the terms of the GNU Lesser General Public License as
// published by the Free Software Foundation, either version 3 of the
// License, or (at your option) any later version.
//
// WANN is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
// or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General
// Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public
// License along with WANN.  If not, see <http://www.gnu.org/licenses/>.
//----------------------------------------------------------------------
// Acknowledgment:
//
// The development of the WANN library has been supported by ANR
// (Agence Nationale de la Recherche, France) under contract 
// BLAN07-2_183172 OTARIE and by RFBR (Russian Foundation for Basic 
// Research) under contract RFBR/CNRS 07-01-92217.
//----------------------------------------------------------------------
// History:
//
// This file has been originally released under GNU LGPL as part of the ANN
// (Approximate Nearest Neighbors) library of David Mount and Sunil
// Arya.  It has been modified to become part of the WANN library.
//
// WANN history:
//	Revision 0.2 06/22/08
//		Array L handled in ANNbruteForce, ~ANNbruteForce.
//	Revision 0.1 06/03/08
//		Modified copyright information.
//		Changes comments.
//		Added argument wa to ANNbruteForce class constructor.
//		Take into account weights in annkSearch.
//
// ANN history:
//	Revision 1.1  05/03/05
//		Added fixed-radius kNN search
//	Revision 0.1  03/04/98
//		Initial release
//----------------------------------------------------------------------

#include <wann/ANNx.h>					// all ANN includes
#include "pr_queue_k.h"					// k element priority queue

//----------------------------------------------------------------------
//		Brute-force search simply stores a pointer to the list of
//		data points and searches linearly for the nearest neighbor.
//		The k nearest neighbors are stored in a k-element priority
//		queue (which is implemented in a pretty dumb way as well).
//
//		If ANN_ALLOW_SELF_MATCH is ANNfalse then data points at distance
//		zero are not considered.
//
//		Note that the error bound eps is passed in, but it is ignored.
//		These routines compute exact nearest neighbors (which is needed
//		for validation purposes in ann_test.cxx).
//----------------------------------------------------------------------

ANNbruteForce::ANNbruteForce(			// constructor from point array
	ANNpointArray		pa,				// point array
#ifndef ANN_NO_WEIGHTS
	ANNdistArray		wa,				// array of weights
#endif
	int					n,				// number of points
#ifdef WANN_TORUS	
    ANNdistArray        la,             // size of the space
#endif	
	int					dd)				// dimension
{
	dim = dd;  n_pts = n;  pts = pa;
#ifndef ANN_NO_WEIGHTS
	weights = wa;
#endif
#ifdef WANN_TORUS
    L = new ANNdist[dim];               //alloc space for    
    L = la;
#endif	
}

ANNbruteForce::~ANNbruteForce() {       // destructor (not empty)
#ifdef WANN_TORUS
    delete [] L;
    L = NULL;
#endif    
}

void ANNbruteForce::annkSearch(			// approx k near neighbor search
	ANNpoint			q,				// query point
	int					k,				// number of near neighbors to return
	ANNidxArray			nn_idx,			// nearest neighbor indices (returned)
	ANNdistArray		dd,				// dist to near neighbors (returned)
	double				eps)			// error bound (ignored)
{
	ANNmin_k mk(k);						// construct a k-limited priority queue
	int i;

	if (k > n_pts) {					// too many near neighbors?
		annError("Requesting more near neighbors than data points", ANNabort);
	}
										// run every point through queue
	for (i = 0; i < n_pts; i++) {
										// compute distance to point
		ANNdist sqDist = annDist(dim, 
#ifdef WANN_TORUS
        L,
#endif        
        pts[i], q);
		if (ANN_ALLOW_SELF_MATCH || sqDist != 0)
#ifndef ANN_NO_WEIGHTS
			mk.insert(sqDist + weights[i], i);
#else
			mk.insert(sqDist, i);
#endif
	}
	for (i = 0; i < k; i++) {			// extract the k closest points
		dd[i] = mk.ith_smallest_key(i);
		nn_idx[i] = mk.ith_smallest_info(i);
	}
}

int ANNbruteForce::annkFRSearch(		// approx fixed-radius kNN search
	ANNpoint			q,				// query point
	ANNdist				sqRad,			// squared radius
	int					k,				// number of near neighbors to return
	ANNidxArray			nn_idx,			// nearest neighbor array (returned)
	ANNdistArray		dd,				// dist to near neighbors (returned)
	double				eps)			// error bound
{
	ANNmin_k mk(k);						// construct a k-limited priority queue
	int i;
	int pts_in_range = 0;				// number of points in query range
										// run every point through queue
	for (i = 0; i < n_pts; i++) {
										// compute distance to point
		ANNdist sqDist = annDist(dim, 
#ifdef WANN_TORUS
        L,
#endif        
        pts[i], q);
		if (sqDist <= sqRad &&			// within radius bound
			(ANN_ALLOW_SELF_MATCH || sqDist != 0)) { // ...and no self match
#ifndef ANN_NO_WEIGHTS
			mk.insert(sqDist + weights[i], i);
#else
			mk.insert(sqDist, i);
#endif
			pts_in_range++;
		}
	}
	for (i = 0; i < k; i++) {			// extract the k closest points
		if (dd != NULL)
			dd[i] = mk.ith_smallest_key(i);
		if (nn_idx != NULL)
			nn_idx[i] = mk.ith_smallest_info(i);
	}

	return pts_in_range;
}
