//----------------------------------------------------------------------
// File:			kd_fix_rad_search.cxx
// Programmer:		Sunil Arya and David Mount,
//					Andrei Sobolevskii and Aleksandr Andrievskii
// Description:		Standard kd-tree fixed-radius kNN search
// Last modified:	06/22/08
//----------------------------------------------------------------------
// Copyright 2008 Andrei Sobolevskii and Aleksandr Andrievskii
//
// Copyright (c) 1997-2005 University of Maryland and Sunil Arya and
// David Mount.  All Rights Reserved.
// 
// This file is part of the WANN (Weighted Approximate Nearest 
// Neighbors) library.
//
// WANN is free software: you can redistribute it and/or modify it
// under the terms of the GNU Lesser General Public License as
// published by the Free Software Foundation, either version 3 of the
// License, or (at your option) any later version.
//
// WANN is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
// or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General
// Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public
// License along with WANN.  If not, see <http://www.gnu.org/licenses/>.
//----------------------------------------------------------------------
// Acknowledgment:
//
// The development of the WANN library has been supported by ANR
// (Agence Nationale de la Recherche, France) under contract 
// BLAN07-2_183172 OTARIE and by RFBR (Russian Foundation for Basic 
// Research) under contract RFBR/CNRS 07-01-92217.
//----------------------------------------------------------------------
// History:
//
// This file has been originally released under GNU LGPL as part of the ANN
// (Approximate Nearest Neighbors) library of David Mount and Sunil
// Arya.  It has been modified to become part of the WANN library.
//
// WANN history:
//	Revision 0.2 06/22/08
//		New global arrays ANNkdFRL (torus sizes), ANNkdFRWeights
//		Added utility functions fr_upd_dist, fr_not_need_inversion
//		Toric geometry in ann_fr_search
//		Fixed handling of weights in leaf nodes
//	Revision 0.1 06/03/08
//		Modified copying permission
//
// ANN history:
//	Revision 1.1  05/03/05
//		Initial release
//----------------------------------------------------------------------

#include "kd_fix_rad_search.h"			// kd fixed-radius search decls

//----------------------------------------------------------------------
//	Approximate fixed-radius k nearest neighbor search
//		The squared radius is provided, and this procedure finds the
//		k nearest neighbors within the radius, and returns the total
//		number of points lying within the radius.
//
//		The method used for searching the kd-tree is a variation of the
//		nearest neighbor search used in kd_search.cxx, except that the
//		radius of the search ball is known.  We refer the reader to that
//		file for the explanation of the recursive search procedure.
//----------------------------------------------------------------------

//----------------------------------------------------------------------
//		To keep argument lists short, a number of global variables
//		are maintained which are common to all the recursive calls.
//		These are given below.
//----------------------------------------------------------------------

int				ANNkdFRDim;				// dimension of space
ANNpoint		ANNkdFRQ;				// query point
ANNdist			ANNkdFRSqRad;			// squared radius search bound
double			ANNkdFRMaxErr;			// max tolerable squared error
ANNpointArray	ANNkdFRPts;				// the points
#ifndef ANN_NO_WEIGHTS
ANNdistArray	ANNkdFRWeights;			// the weights
#endif
ANNmin_k*		ANNkdFRPointMK;			// set of k closest points
int				ANNkdFRPtsVisited;		// total points visited
int				ANNkdFRPtsInRange;		// number of points in the range

#ifdef WANN_TORUS
ANNdistArray    ANNkdFRL;  //size of the space, copied from the tree
#endif

//----------------------------------------------------------------------
//	annkFRSearch - fixed radius search for k nearest neighbors
//----------------------------------------------------------------------

int ANNkd_tree::annkFRSearch(
	ANNpoint			q,				// the query point
	ANNdist				sqRad,			// squared radius search bound
	int					k,				// number of near neighbors to return
	ANNidxArray			nn_idx,			// nearest neighbor indices (returned)
	ANNdistArray		dd,				// the approximate nearest neighbor
	double				eps)			// the error bound
{
	ANNkdFRDim = dim;					// copy arguments to static equivs
	ANNkdFRQ = q;
	ANNkdFRSqRad = sqRad;
	ANNkdFRPts = pts;
#ifndef ANN_NO_WEIGHTS
	ANNkdFRWeights = weights;
#endif
#ifdef WANN_TORUS
    ANNkdFRL = L;
#endif
	ANNkdFRPtsVisited = 0;				// initialize count of points visited
	ANNkdFRPtsInRange = 0;				// ...and points in the range

	ANNkdFRMaxErr = ANN_POW(1.0 + eps);
	ANN_FLOP(2)							// increment floating op count

	ANNkdFRPointMK = new ANNmin_k(k);	// create set for closest k points
										// search starting at the root
	root->ann_FR_search(annBoxDistance(q, bnd_box_lo, bnd_box_hi, dim
#ifdef WANN_TORUS
    ,L
#endif
    )
#ifdef WANN_TORUS
       ,q
#endif    
    );

	for (int i = 0; i < k; i++) {		// extract the k-th closest points
		if (dd != NULL)
			dd[i] = ANNkdFRPointMK->ith_smallest_key(i);
		if (nn_idx != NULL)
			nn_idx[i] = ANNkdFRPointMK->ith_smallest_info(i);
	}

	delete ANNkdFRPointMK;				// deallocate closest point set
	return ANNkdFRPtsInRange;			// return final point count
}

#ifdef WANN_TORUS
//-------------------------------------------------------------------------
// torus search utilites - some functions to simplify code
//-------------------------------------------------------------------------

//simple replacement 
ANNdist fr_upd_dist(ANNdist old_dist,
                 ANNdist old_diff,
                 ANNdist new_diff){
    return (ANNdist) ANN_SUM(old_dist,
				ANN_DIFF(ANN_POW(old_diff), ANN_POW(new_diff)));    
}        

ANNbool fr_not_need_inversion(ANNcoord point,
                        ANNcoord btm,
                        ANNcoord top,
                        ANNidx cut_dim){
    //will the inversion decrease the distance to the node?
    if (point < btm) {
        if (point + ANNkdFRL[cut_dim] - top > btm - point)
			return ANNtrue;
		else return ANNfalse;
	} else if (point >top) {
		if (btm - point + ANNkdFRL[cut_dim] > point - top)
			return ANNtrue;
		else return ANNfalse;
	} else return ANNtrue; //point in the box - no inversion    
}
#endif

//----------------------------------------------------------------------
//	kd_split::ann_FR_search - search a splitting node
//		Note: This routine is similar in structure to the standard kNN
//		search.  It visits the subtree that is closer to the query point
//		first.  For fixed-radius search, there is no benefit in visiting
//		one subtree before the other, but we maintain the same basic
//		code structure for the sake of uniformity.
//----------------------------------------------------------------------

void ANNkd_split::ann_FR_search(ANNdist box_dist
#ifdef WANN_TORUS
       ,ANNpoint			q
#endif
)
{
										// check dist calc term condition
	if (ANNmaxPtsVisited != 0 && ANNkdFRPtsVisited > ANNmaxPtsVisited) return;

#ifdef WANN_TORUS
	// in toroidal case ANNkdQ holds the true query point, 
	// while particular nodes may find it more fruitful to work
	// with their own custom point inverted along one or more axis

    ANNcoord cut_diff = q[cut_dim] - cut_val;
    ANNidx first_child;		// first child node to visit
    ANNidx second_child;	// second child node to visit
    ANNdist first_dist;		// distance to the first node
    ANNdist second_dist;	// distance to the node child[second_child]
    ANNpoint first_point;	// point to be passed to the first node
    ANNpoint second_point;	// for the second one    

    //points that will be passed to LO and HI nodes 
    //(may be either first or second!)
    ANNpoint lo_point = annCopyPt(ANNkdFRDim, q);
    ANNpoint hi_point = annCopyPt(ANNkdFRDim, q);
    ANNdist lo_dist = box_dist;
    ANNdist hi_dist = box_dist; 

    //three choices that determine where to go

    //first choice: do we need to invert lo_point
    if (fr_not_need_inversion(q[cut_dim], cd_bnds[ANN_LO], cut_val, cut_dim)) {
        //lo_point: no changes
        if (cut_diff > 0) { //only in this case we need to update distance
    		ANNcoord box_diff = lo_point[cut_dim] - cd_bnds[ANN_HI];
	    	if (box_diff < 0)				// within bounds - ignore
		    	box_diff = 0;
		    lo_dist = fr_upd_dist(box_dist, box_diff, cut_diff);
        }		
    } else {
        //invert lo_point
        if (cut_diff < 0) { 
            lo_point[cut_dim] = lo_point[cut_dim] + ANNkdFRL[cut_dim];
     		ANNcoord box_diff = cd_bnds[ANN_LO] - q[cut_dim];
	    	if (box_diff < 0) {				// within bounds - ignore
		    	box_diff = 0; // cannot be in lo box		    	
		    	cerr << "lo_point was in lo_box before inversion!!!" << endl;
             }	
		    lo_dist = fr_upd_dist(box_dist, box_diff, lo_point[cut_dim] - cut_val);
        } else {
            lo_point[cut_dim] = lo_point[cut_dim] - ANNkdFRL[cut_dim];
     		ANNcoord box_diff = q[cut_dim] - cd_bnds[ANN_HI];
	    	if (box_diff < 0)				// within bounds - ignore
		    	box_diff = 0; //here we may be in the upper box - if it is large enough
		    lo_dist = fr_upd_dist(box_dist, box_diff, cd_bnds[ANN_LO] - lo_point[cut_dim]);	
        }    
    }          
    //second choice: do we need to invert hi_point?
    //much like the prevoius one except for replaces like hi<-lo and dependent
        if (fr_not_need_inversion(q[cut_dim], cut_val, cd_bnds[ANN_HI], cut_dim)) {
        //hi_point: no changes
        if (cut_diff < 0) { //only in this case we need to update distance
    		ANNcoord box_diff = cd_bnds[ANN_LO] - hi_point[cut_dim];
	    	if (box_diff < 0)				// within bounds - ignore
		    	box_diff = 0;
		    hi_dist = fr_upd_dist(box_dist, box_diff, cut_diff);
        }		
    } else {
        //invert hi_point
        if (cut_diff > 0) { 
            hi_point[cut_dim] = hi_point[cut_dim] - ANNkdFRL[cut_dim];
     		ANNcoord box_diff = q[cut_dim] - cd_bnds[ANN_HI];
	    	if (box_diff < 0) {				// within bounds - ignore
		    	box_diff = 0; // cannot be in hi box because inner points are not inverted
		    	cerr << "hi_point was in hi_box before inversion!!!" << endl;
             }	
		    hi_dist = fr_upd_dist(box_dist, box_diff, cut_val - hi_point[cut_dim]);
        } else {
            hi_point[cut_dim] = hi_point[cut_dim] + ANNkdFRL[cut_dim];
     		ANNcoord box_diff = cd_bnds[ANN_LO] - q[cut_dim];
	    	if (box_diff < 0)				// within bounds - ignore
		    	box_diff = 0; //here we may be in the upper box - if it is large enough
		    hi_dist = fr_upd_dist(box_dist, box_diff, hi_point[cut_dim] - cd_bnds[ANN_HI]);
        }    
    }

    //third choice: where do you want to go today?
    if (lo_dist < hi_dist){// condition compatible with ANN_WEIGHTS 
        first_child   = ANN_LO;
        first_point   = lo_point;
        first_dist    = lo_dist;
        second_child = ANN_HI;
        second_point = hi_point;
        second_dist  = hi_dist;        
    } else {
        first_child   = ANN_HI;
        first_point   = hi_point;
        first_dist    = hi_dist;
        second_child = ANN_LO;
        second_point = lo_point;
        second_dist  = lo_dist;                  
    }
    child[first_child]->ann_FR_search(first_dist, first_point);   
    //check that we should visit another child
    if(second_dist * ANNkdFRMaxErr <= ANNkdFRSqRad){
        child[second_child]->ann_FR_search(second_dist, second_point);
}
    first_point = NULL;
    second_point = NULL;
    delete[] lo_point;
    delete[] hi_point;
#else // !WANN_TORUS
										// distance to cutting plane
	ANNcoord cut_diff = ANNkdFRQ[cut_dim] - cut_val;

	if (cut_diff < 0) {					// left of cutting plane
		child[ANN_LO]->ann_FR_search(box_dist);// visit closer child first

		ANNcoord box_diff = cd_bnds[ANN_LO] - ANNkdFRQ[cut_dim];
		if (box_diff < 0)				// within bounds - ignore
			box_diff = 0;
										// distance to further box
		box_dist = (ANNdist) ANN_SUM(box_dist,
				ANN_DIFF(ANN_POW(box_diff), ANN_POW(cut_diff)));

										// visit further child if in range
		if (box_dist * ANNkdFRMaxErr <= ANNkdFRSqRad)
			child[ANN_HI]->ann_FR_search(box_dist);

	}
	else {								// right of cutting plane
		child[ANN_HI]->ann_FR_search(box_dist);// visit closer child first

		ANNcoord box_diff = ANNkdFRQ[cut_dim] - cd_bnds[ANN_HI];
		if (box_diff < 0)				// within bounds - ignore
			box_diff = 0;
										// distance to further box
		box_dist = (ANNdist) ANN_SUM(box_dist,
				ANN_DIFF(ANN_POW(box_diff), ANN_POW(cut_diff)));

										// visit further child if close enough
		if (box_dist * ANNkdFRMaxErr <= ANNkdFRSqRad)
			child[ANN_LO]->ann_FR_search(box_dist);

	}
	ANN_FLOP(13)						// increment floating ops
	ANN_SPL(1)							// one more splitting node visited
#endif // !WANN_TORUS
}

//----------------------------------------------------------------------
//	kd_leaf::ann_FR_search - search points in a leaf node
//		Note: The unreadability of this code is the result of
//		some fine tuning to replace indexing by pointer operations.
//----------------------------------------------------------------------

void ANNkd_leaf::ann_FR_search(ANNdist box_dist
#ifdef WANN_TORUS
							   , ANNpoint q
#endif
)
{
	register ANNdist dist;				// distance to data point
	register ANNcoord* pp;				// data coordinate pointer
	register ANNcoord* qq;				// query coordinate pointer
	register ANNcoord t;
	register int d;

	for (int i = 0; i < n_pts; i++) {	// check points in bucket

		pp = ANNkdFRPts[bkt[i]];		// first coord of next data point
		qq = ANNkdFRQ;					// first coord of query point
		dist = 0;

		for(d = 0; d < ANNkdFRDim; d++) {
			ANN_COORD(1)				// one more coordinate hit
			ANN_FLOP(5)					// increment floating ops

			t = *(qq++) - *(pp++);		// compute length and adv coordinate
										// exceeds dist to k-th smallest?

// WANN_TORUS: here we use true points instead of (probably) inverted
// ones.  The reason is that given a large enough bucket size, one can
// encounter both points that require inversion and not.
#ifdef WANN_TORUS	
			if (t > ANNkdFRL[d]/2)		// the true distance in toroidal space
				t = t - ANNkdFRL[d];
			else if (t < -ANNkdFRL[d]/2)
				t = t + ANNkdFRL[d];
#endif
			if ((dist = ANN_SUM(dist, ANN_POW(t))) > ANNkdFRSqRad)
				break;
		}

		if (d >= ANNkdFRDim &&					// among the k best?
		   (ANN_ALLOW_SELF_MATCH || dist!=0)) { // and no self-match problem
												// add it to the list
#ifndef ANN_NO_WEIGHTS
			ANNkdFRPointMK->insert(dist + ANNkdFRWeights[bkt[i]], bkt[i]);
#else
			ANNkdFRPointMK->insert(dist, bkt[i]);
#endif
			ANNkdFRPtsInRange++;				// increment point count
		}
	}
	ANN_LEAF(1)							// one more leaf node visited
	ANN_PTS(n_pts)						// increment points visited
	ANNkdFRPtsVisited += n_pts;			// increment number of points visited
}
