//----------------------------------------------------------------------
// File:			kd_pr_search.cxx
// Programmer:		Sunil Arya and David Mount,
//					Andrei Sobolevskii and Aleksandr Andrievskii
// Description:		Priority search for kd-trees
// Last modified:	06/22/08
//----------------------------------------------------------------------
// Copyright 2008 Andrei Sobolevskii and Aleksandr Andrievskii
//
// Copyright (c) 1997-2005 University of Maryland and Sunil Arya and
// David Mount.  All Rights Reserved.
// 
// This file is part of the WANN (Weighted Approximate Nearest 
// Neighbors) library.
//
// WANN is free software: you can redistribute it and/or modify it
// under the terms of the GNU Lesser General Public License as
// published by the Free Software Foundation, either version 3 of the
// License, or (at your option) any later version.
//
// WANN is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
// or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General
// Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public
// License along with WANN.  If not, see <http://www.gnu.org/licenses/>.
//----------------------------------------------------------------------
// Acknowledgment:
//
// The development of the WANN library has been supported by ANR
// (Agence Nationale de la Recherche, France) under contract 
// BLAN07-2_183172 OTARIE and by RFBR (Russian Foundation for Basic 
// Research) under contract RFBR/CNRS 07-01-92217.
//----------------------------------------------------------------------
// History:
//
// This file has been originally released under GNU LGPL as part of the ANN
// (Approximate Nearest Neighbors) library of David Mount and Sunil
// Arya.  It has been modified to become part of the WANN library.
//
// WANN history:
//	Revision 0.2 06/22/08
//		New global array ANNkdFRL (torus sizes)
//		Added utility functions pr_upd_dist, pr_not_need_inversion
//		Toric geometry in ann_pri_search
//	Revision 0.1 06/03/08
//		Modified copyright information.
//		Changes comments.
//		Added global array ANNprWeights.
//		Modified ANNkd_tree::annkPriSearch.
//		Treat low anf high boxes separately in ANNkd_split::ann_pri_search.
//		Add weight to distances in ANNkd_leaf::ann_pr_search.
//		
//		
// ANN history:
//	Revision 0.1  03/04/98
//		Initial release
//----------------------------------------------------------------------

#include "kd_pr_search.h"				// kd priority search declarations

//----------------------------------------------------------------------
//	Approximate nearest neighbor searching by priority search.
//		The kd-tree is searched for an approximate nearest neighbor.
//		The point is returned through one of the arguments, and the
//		distance returned is the SQUARED distance to this point.
//
//		The method used for searching the kd-tree is called priority
//		search.  (It is described in Arya and Mount, ``Algorithms for
//		fast vector quantization,'' Proc. of DCC '93: Data Compression
//		Conference}, eds. J. A. Storer and M. Cohn, IEEE Press, 1993,
//		381--390.)
//
//		The cell of the kd-tree containing the query point is located,
//		and cells are visited in increasing order of distance from the
//		query point.  This is done by placing each subtree which has
//		NOT been visited in a priority queue, according to the closest
//		distance of the corresponding enclosing rectangle from the
//		query point.  The search stops when the distance to the nearest
//		remaining rectangle exceeds the distance to the nearest point
//		seen by a factor of more than 1/(1+eps). (Implying that any
//		point found subsequently in the search cannot be closer by more
//		than this factor.)
//
//		The main entry point is annkPriSearch() which sets things up and
//		then call the recursive routine ann_pri_search().  This is a
//		recursive routine which performs the processing for one node in
//		the kd-tree.  There are two versions of this virtual procedure,
//		one for splitting nodes and one for leaves. When a splitting node
//		is visited, we determine which child to continue the search on
//		(the closer one), and insert the other child into the priority
//		queue.  When a leaf is visited, we compute the distances to the
//		points in the buckets, and update information on the closest
//		points.
//
//		Some trickery is used to incrementally update the distance from
//		a kd-tree rectangle to the query point.  This comes about from
//		the fact that which each successive split, only one component
//		(along the dimension that is split) of the squared distance to
//		the child rectangle is different from the squared distance to
//		the parent rectangle.
//----------------------------------------------------------------------

//----------------------------------------------------------------------
//		To keep argument lists short, a number of global variables
//		are maintained which are common to all the recursive calls.
//		These are given below.
//----------------------------------------------------------------------

double			ANNprEps;				// the error bound
int				ANNprDim;				// dimension of space
ANNpoint		ANNprQ;					// query point
double			ANNprMaxErr;			// max tolerable squared error
ANNpointArray	ANNprPts;				// the points
#ifndef ANN_NO_WEIGHTS
ANNdistArray	ANNprWeights;			// the weights
#endif
ANNpr_queue		*ANNprBoxPQ;			// priority queue for boxes
ANNmin_k		*ANNprPointMK;			// set of k closest points

#ifdef WANN_TORUS
ANNdistArray    ANNprL;  //size of the space, copied from the tree
ANNpr_queue		*ANNprPointPQ;			// priority queue for boxes
#endif

//----------------------------------------------------------------------
//	annkPriSearch - priority search for k nearest neighbors
//----------------------------------------------------------------------

void ANNkd_tree::annkPriSearch(
	ANNpoint			q,				// query point
	int					k,				// number of near neighbors to return
	ANNidxArray			nn_idx,			// nearest neighbor indices (returned)
	ANNdistArray		dd,				// dist to near neighbors (returned)
	double				eps)			// error bound (ignored)
{
										// max tolerable squared error
	ANNprMaxErr = ANN_POW(1.0 + eps);
	ANN_FLOP(2)							// increment floating ops

	ANNprDim = dim;						// copy arguments to static equivs
	ANNprQ = q;
	ANNprPts = pts;
#ifndef ANN_NO_WEIGHTS
	ANNprWeights = weights;
#endif
#ifdef WANN_TORUS
    ANNprL = L;
#endif
	ANNptsVisited = 0;					// initialize count of points visited

	ANNprPointMK = new ANNmin_k(k);		// create set for closest k points

										// distance to root box
	ANNdist box_dist = annBoxDistance(q,
				bnd_box_lo, bnd_box_hi, dim
#ifdef WANN_TORUS
    , L
#endif
                );

	ANNprBoxPQ = new ANNpr_queue(n_pts);	// create priority queue for boxes
	ANNprBoxPQ->insert(box_dist, root);		// insert root in priority queue
#ifdef WANN_TORUS	
    ANNprPointPQ = new ANNpr_queue(n_pts);	// create priority queue for points
	ANNpoint p = annCopyPt(ANNprDim, q);
    ANNprPointPQ->insert(box_dist, p);		// insert non-inverted point
#endif

	while (ANNprBoxPQ->non_empty() &&
		(!(ANNmaxPtsVisited != 0 && ANNptsVisited > ANNmaxPtsVisited))) {
		ANNkd_ptr np;					// next box from prior queue

										// extract closest box from queue
		ANNprBoxPQ->extr_min(box_dist, (void *&) np);
#ifdef WANN_TORUS	
		ANNprPointPQ->extr_min(box_dist, (void *&)p); //closest point
#endif

#ifndef ANN_NO_WEIGHTS
		ANN_FLOP(3)
			if((box_dist + np->minWeight)*ANNprMaxErr >=
			   ANNprPointMK->max_key())
				break;
#else
		ANN_FLOP(2)						// increment floating ops
		if (box_dist*ANNprMaxErr >= ANNprPointMK->max_key())
			break;
#endif

		np->ann_pri_search(box_dist
#ifdef WANN_TORUS
       , p
#endif    
        );	// search this subtree.
	}

	for (int i = 0; i < k; i++) {		// extract the k-th closest points
		dd[i] = ANNprPointMK->ith_smallest_key(i);
		nn_idx[i] = ANNprPointMK->ith_smallest_info(i);
	}

	delete ANNprPointMK;				// deallocate closest point set
	delete ANNprBoxPQ;					// deallocate priority queue
#ifdef WANN_TORUS	
	delete ANNprPointPQ;
	delete [] p;
#endif    
}

#ifdef WANN_TORUS
//-------------------------------------------------------------------------
// torus search utilites - some functions to simplify code
//-------------------------------------------------------------------------

//simple replacement 
ANNdist pr_upd_dist(ANNdist old_dist,
                 ANNdist old_diff,
                 ANNdist new_diff){
    return (ANNdist) ANN_SUM(old_dist,
				ANN_DIFF(ANN_POW(old_diff), ANN_POW(new_diff)));    
}        

ANNbool pr_not_need_inversion(ANNcoord point,
                        ANNcoord btm,
                        ANNcoord top,
                        ANNidx cut_dim){
    //will the inversion decrease the distance to the node?
    if (point < btm) {
		if (point+ANNprL[cut_dim] - top>btm - point)
			return ANNtrue;
		else return ANNfalse;
	} else if (point >top) {
		if (btm - point+ ANNprL[cut_dim]> point - top)
			return ANNtrue;
		else return ANNfalse;
	} else return ANNtrue; //point in the box - no inversion    
}
#endif

//----------------------------------------------------------------------
//	kd_split::ann_pri_search - search a splitting node
//----------------------------------------------------------------------

void ANNkd_split::ann_pri_search(ANNdist box_dist
#ifdef WANN_TORUS
								 , ANNpoint q
#endif
)
{
#ifndef ANN_NO_WEIGHTS
	ANNdist lo_box_dist, hi_box_dist;
#else
	ANNdist new_dist;					// distance to child visited later
#endif

#ifdef ANN_TORUS   
// in toroidal case ANNkdQ holds the true query point, 
// while particular nodes may find it more fruitful to work
// with their own custom point inverted along one or more axis
    ANNcoord cut_diff = q[cut_dim] - cut_val;
    ANNidx first_child;		// first child node to visit
    ANNidx second_child;	// second child node to visit
    ANNdist first_dist;		// distance to the first node
    ANNdist second_dist;	// distance to the node child[second_child]
    ANNpoint first_point;	// point to be passed to the first node
    ANNpoint second_point;	// for the second one

    // points that will be passed to LO and HI nodes 
    // (may be either first or second!)
    ANNpoint lo_point = annCopyPt(ANNprDim, q);
    ANNpoint hi_point = annCopyPt(ANNprDim, q);
    ANNdist lo_dist = box_dist;
    ANNdist hi_dist = box_dist; 

    cerr << "new node cd =" << cut_dim << ", dist =" << box_dist;
    cerr << ",  point =" << q[cut_dim] << endl;

    // three choices that determine where to go
    // first choice: do we need to invert lo_point?
    if (pr_not_need_inversion(q[cut_dim], cd_bnds[ANN_LO], cut_val, cut_dim)) {
        // lo_point: no changes
        if (cut_diff>0) { //only in this case we need to update distance
    		ANNcoord box_diff = lo_point[cut_dim] - cd_bnds[ANN_HI];
	    	if (box_diff < 0)				// within bounds - ignore
		    	box_diff = 0;
		    lo_dist = pr_upd_dist(box_dist, box_diff, cut_diff);
        }		
    } else {
        //invert lo_point
        if (cut_diff<0) { 
            lo_point[cut_dim] = lo_point[cut_dim] + ANNprL[cut_dim];
     		ANNcoord box_diff = cd_bnds[ANN_LO] - q[cut_dim];
	    	if (box_diff < 0) {				// within bounds - ignore
		    	box_diff = 0; // cannot be in lo box		    	
		    	cerr << "lo_point was in lo_box before inversion!!!" << endl;
             }	
		    lo_dist = pr_upd_dist(box_dist, box_diff, lo_point[cut_dim] - cut_val);
        } else {
            lo_point[cut_dim] = lo_point[cut_dim] - ANNprL[cut_dim];
     		ANNcoord box_diff = q[cut_dim] - cd_bnds[ANN_HI];
	    	if (box_diff < 0)				// within bounds - ignore
		    	box_diff = 0; // here we may be in the upper box - if it is large enough 
		    lo_dist = pr_upd_dist(box_dist, box_diff, cd_bnds[ANN_LO] - lo_point[cut_dim]);	
        }    
    }          
    //second one: do we need to invert hi_point?
    //much like the prevoius choice excepot for replaces like hi<-lo and dependent
        if (pr_not_need_inversion(q[cut_dim], cut_val, cd_bnds[ANN_HI], cut_dim)) {
        //hi_point: no changes
        if (cut_diff < 0) { //only in this case we need to update distance
    		ANNcoord box_diff = cd_bnds[ANN_LO] - hi_point[cut_dim];
	    	if (box_diff < 0)				// within bounds - ignore
		    	box_diff = 0;
		    hi_dist = pr_upd_dist(box_dist, box_diff, cut_diff);
        }		
    } else {
        //invert hi_point
        if (cut_diff > 0) { 
            hi_point[cut_dim] = hi_point[cut_dim] - ANNprL[cut_dim];
     		ANNcoord box_diff = q[cut_dim] - cd_bnds[ANN_HI];
	    	if (box_diff < 0) {				// within bounds - ignore
		    	box_diff = 0; // cannot be in hi box because inner points are not inverted
		    	cerr << "hi_point was in hi_box before inversion!!!" << endl;
             }	
		    hi_dist = pr_upd_dist(box_dist, box_diff, cut_val - hi_point[cut_dim]);
        } else {
            hi_point[cut_dim] = hi_point[cut_dim] + ANNprL[cut_dim];
     		ANNcoord box_diff = cd_bnds[ANN_LO] - q[cut_dim];
	    	if (box_diff < 0)				// within bounds - ignore
		    	box_diff = 0; // here we may be in the upper box - if it is large enough
		    hi_dist = pr_upd_dist(box_dist, box_diff, hi_point[cut_dim] - cd_bnds[ANN_HI]);
        }    
    }          
    //third choice: where do you want to go today?
#ifndef ANN_NO_WEIGHTS
	if (lo_dist + child[ANN_LO]->minHeight <
		hi_dist + child[ANN_HI]->minHeight)
#else
    if (lo_dist<hi_dist)
#endif    
        { // low child first
        first_child   = ANN_LO;
        first_point   = lo_point;
        first_dist    = lo_dist;
        second_child = ANN_HI;
        second_point = hi_point;
        second_dist  = hi_dist;        
    } else {
        first_child   = ANN_HI;
        first_point   = hi_point;
        first_dist    = hi_dist;
        second_child = ANN_LO;
        second_point = lo_point;
        second_dist  = lo_dist;                  
    }
   if (child[second_child] != KD_TRIVIAL) { // enqueue if not trivial
   cerr << " store: cd = " << cut_dim << ", point =" << second_point[cut_dim] << endl;
			ANNprBoxPQ->insert(second_dist, child[second_child]);
			ANNprPointPQ->insert(second_dist, second_point);
   }		
    child[first_child]->ann_pri_search(first_dist, first_point);   

    first_point = NULL;
    if (first_child == ANN_LO)
        delete[] lo_point; // this point (and node) no longer needed
    else    
        delete[] hi_point;
#else // !WANN_TORUS: the flat open space
										// distance to cutting plane
	ANNcoord cut_diff = ANNprQ[cut_dim] - cut_val;

#ifndef ANN_NO_WEIGHTS
	if (cut_diff < 0) {					// left of cutting plane
		lo_box_dist = box_dist;			// low box closer
										// find dist to farther box
		ANNcoord box_diff = cd_bnds[ANN_LO] - ANNprQ[cut_dim];
		if (box_diff < 0)				// within bounds - ignore
			box_diff = 0;
										// distance to farther box
		hi_box_dist = (ANNdist) ANN_SUM(box_dist,
				ANN_DIFF(ANN_POW(box_diff), ANN_POW(cut_diff)));
	}
	else {								// right of cutting plane
		hi_box_dist = box_dist;			// high box closer
										// find dist to farther box
		ANNcoord box_diff = ANNprQ[cut_dim] - cd_bnds[ANN_HI];
		if (box_diff < 0)				// within bounds - ignore
			box_diff = 0;
										// distance to further box
		lo_box_dist = (ANNdist) ANN_SUM(box_dist,
				ANN_DIFF(ANN_POW(box_diff), ANN_POW(cut_diff)));
	}

	if (lo_box_dist + child[ANN_LO]->minWeight <
		hi_box_dist + child[ANN_HI]->minWeight)
	{									// low child first
		if (child[ANN_HI] != KD_TRIVIAL)// enqueue if not trivial
			ANNprBoxPQ->insert(hi_box_dist, child[ANN_HI]);
										// continue with better child
		child[ANN_LO]->ann_pri_search(lo_box_dist);
	}
	else {
		if (child[ANN_LO] != KD_TRIVIAL)// enqueue if not trivial
			ANNprBoxPQ->insert(lo_box_dist, child[ANN_LO]);
										// continue with closer child
		child[ANN_HI]->ann_pri_search(hi_box_dist);
	}
	ANN_FLOP(3)
#else // ANN_NO_WEIGHTS
	if (cut_diff < 0) {					// left of cutting plane
		ANNcoord box_diff = cd_bnds[ANN_LO] - ANNprQ[cut_dim];
		if (box_diff < 0)				// within bounds - ignore
			box_diff = 0;
										// distance to further box
		new_dist = (ANNdist) ANN_SUM(box_dist,
				ANN_DIFF(ANN_POW(box_diff), ANN_POW(cut_diff)));

		if (child[ANN_HI] != KD_TRIVIAL)// enqueue if not trivial
			ANNprBoxPQ->insert(new_dist, child[ANN_HI]);
										// continue with closer child
		child[ANN_LO]->ann_pri_search(box_dist);
	}
	else {								// right of cutting plane
		ANNcoord box_diff = ANNprQ[cut_dim] - cd_bnds[ANN_HI];
		if (box_diff < 0)				// within bounds - ignore
			box_diff = 0;
										// distance to further box
		new_dist = (ANNdist) ANN_SUM(box_dist,
				ANN_DIFF(ANN_POW(box_diff), ANN_POW(cut_diff)));

		if (child[ANN_LO] != KD_TRIVIAL)// enqueue if not trivial
			ANNprBoxPQ->insert(new_dist, child[ANN_LO]);
										// continue with closer child
		child[ANN_HI]->ann_pri_search(box_dist);
	}
#endif // ANN_NO_WEIGHTS
#endif // !WANN_TORUS
	ANN_SPL(1)							// one more splitting node visited
	ANN_FLOP(8)							// increment floating ops
}

//----------------------------------------------------------------------
//	kd_leaf::ann_pri_search - search points in a leaf node
//
//		This is virtually identical to the ann_search for standard search.
//----------------------------------------------------------------------

void ANNkd_leaf::ann_pri_search(ANNdist box_dist
#ifdef WANN_TORUS
       ,ANNpoint			q
#endif
)
{
	register ANNdist dist;				// distance to data point
	register ANNcoord* pp;				// data coordinate pointer
	register ANNcoord* qq;				// query coordinate pointer
	register ANNdist min_dist;			// distance to k-th closest point
	register ANNcoord t;
	register int d;

	min_dist = ANNprPointMK->max_key(); // k-th smallest distance so far

	for (int i = 0; i < n_pts; i++) {	// check points in bucket

		pp = ANNprPts[bkt[i]];			// first coord of next data point
		qq = ANNprQ;					// first coord of query point

#ifndef ANN_NO_WEIGHTS
		dist = ANNprWeights[bkt[i]];
#else
		dist = 0;
#endif

		for(d = 0; d < ANNprDim; d++) {
			ANN_COORD(1)				// one more coordinate hit
			ANN_FLOP(4)					// increment floating ops

			t = *(qq++) - *(pp++);		// compute length and adv coordinate
										// exceeds dist to k-th smallest?

// WANN_TORUS: here we use true points instead of (probably) inverted
// ones the reason is that given a large enough bucket size, one can
// encounter both points that require inversion and not

#ifdef WANN_TORUS	
            if(t>ANNprL[d]/2)			// the true distance in toroidal space
                t = t - ANNprL[d];
            else if(t<-ANNprL[d]/2)    
                t = t + ANNprL[d];
#endif		
			if ((dist = ANN_SUM(dist, ANN_POW(t))) > min_dist) {
				break;
			}
		}

		if (d >= ANNprDim &&					// among the k best?
		   (ANN_ALLOW_SELF_MATCH || dist!=0)) { // and no self-match problem
												// add it to the list
			ANNprPointMK->insert(dist, bkt[i]);
			min_dist = ANNprPointMK->max_key();
		}
	}
	ANN_LEAF(1)							// one more leaf node visited
	ANN_PTS(n_pts)						// increment points visited
	ANNptsVisited += n_pts;				// increment number of points visited
}
