//----------------------------------------------------------------------
// File:			kd_search.cxx
// Programmer:		Sunil Arya and David Mount,
//					Andrei Sobolevskii and Aleksandr Andrievskii
// Description:		Standard kd-tree search
// Last modified:	06/22/08
//----------------------------------------------------------------------
// Copyright 2008 Andrei Sobolevskii and Aleksandr Andrievskii
//
// Copyright (c) 1997-2005 University of Maryland and Sunil Arya and
// David Mount.  All Rights Reserved.
// 
// This file is part of the WANN (Weighted Approximate Nearest 
// Neighbors) library.
//
// WANN is free software: you can redistribute it and/or modify it
// under the terms of the GNU Lesser General Public License as
// published by the Free Software Foundation, either version 3 of the
// License, or (at your option) any later version.
//
// WANN is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
// or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General
// Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public
// License along with WANN.  If not, see <http://www.gnu.org/licenses/>.
//----------------------------------------------------------------------
// Acknowledgment:
//
// The development of the WANN library has been supported by ANR
// (Agence Nationale de la Recherche, France) under contract 
// BLAN07-2_183172 OTARIE and by RFBR (Russian Foundation for Basic 
// Research) under contract RFBR/CNRS 07-01-92217.
//----------------------------------------------------------------------
// History:
//
// This file has been originally released under GNU LGPL as part of the ANN
// (Approximate Nearest Neighbors) library of David Mount and Sunil
// Arya.  It has been modified to become part of the WANN library.
//
// WANN history:
//	Revision 0.2 06/22/08
//		New global array ANNkdFRL (torus sizes)
//		New utility functions upd_dist, not_need_inversion
//		Toric geometry in ann_search
//	Revision 0.1 06/03/08
//		Modified copyright information.
//		Changes comments.
//		Added global array ANNkdWeights.
//		Track squared distances to low and high children
//			in ANNkd_split::ann_search.
//		Add min weight to squared distances in ANNkd_leaf::ann_search.
//		Modified FLOP count in ANNkd_split::ann_search.
//		
// ANN history:
//	Revision 0.1  03/04/98
//		Initial release
//	Revision 1.0  04/01/05
//		Changed names LO, HI to ANN_LO, ANN_HI
//----------------------------------------------------------------------

#include "kd_search.h"					// kd-search declarations

//----------------------------------------------------------------------
//	Approximate nearest neighbor searching by kd-tree search
//		The kd-tree is searched for an approximate nearest neighbor.
//		The point is returned through one of the arguments, and the
//		distance returned is the squared distance to this point.
//
//		The method used for searching the kd-tree is an approximate
//		adaptation of the search algorithm described by Friedman,
//		Bentley, and Finkel, ``An algorithm for finding best matches
//		in logarithmic expected time,'' ACM Transactions on Mathematical
//		Software, 3(3):209-226, 1977).
//
//		The algorithm operates recursively.  When first encountering a
//		node of the kd-tree we first visit the child which is closest to
//		the query point.  On return, we decide whether we want to visit
//		the other child.  If the box containing the other child exceeds
//		1/(1+eps) times the current best distance, then we skip it (since
//		any point found in this child cannot be closer to the query point
//		by more than this factor.)  Otherwise, we visit it recursively.
//		The distance between a box and the query point is computed exactly
//		(not approximated as is often done in kd-tree), using incremental
//		distance updates, as described by Arya and Mount in ``Algorithms
//		for fast vector quantization,'' Proc.  of DCC '93: Data Compression
//		Conference, eds. J. A. Storer and M. Cohn, IEEE Press, 1993,
//		381-390.
//
//		The main entry points is annkSearch() which sets things up and
//		then call the recursive routine ann_search().  This is a recursive
//		routine which performs the processing for one node in the kd-tree.
//		There are two versions of this virtual procedure, one for splitting
//		nodes and one for leaves.  When a splitting node is visited, we
//		determine which child to visit first (the closer one), and visit
//		the other child on return.  When a leaf is visited, we compute
//		the distances to the points in the buckets, and update information
//		on the closest points.
//
//		Some trickery is used to incrementally update the distance from
//		a kd-tree rectangle to the query point.  This comes about from
//		the fact that which each successive split, only one component
//		(along the dimension that is split) of the squared distance to
//		the child rectangle is different from the squared distance to
//		the parent rectangle.
//----------------------------------------------------------------------

//----------------------------------------------------------------------
//		To keep argument lists short, a number of global variables
//		are maintained which are common to all the recursive calls.
//		These are given below.
//----------------------------------------------------------------------

int				ANNkdDim;				// dimension of space
ANNpoint		ANNkdQ;					// query point
double			ANNkdMaxErr;			// max tolerable squared error
ANNpointArray	ANNkdPts;				// the points
#ifndef ANN_NO_WEIGHTS
ANNdistArray	ANNkdWeights;			// the weights
#endif
ANNmin_k		*ANNkdPointMK;			// set of k closest points
#ifdef WANN_TORUS
ANNdistArray    ANNkdL;  //size of the space, copied from the tree
#endif

//----------------------------------------------------------------------
//	annkSearch - search for the k nearest neighbors
//----------------------------------------------------------------------

void ANNkd_tree::annkSearch(
	ANNpoint			q,				// the query point
	int					k,				// number of near neighbors to return
	ANNidxArray			nn_idx,			// nearest neighbor indices (returned)
	ANNdistArray		dd,				// the approximate nearest neighbor
	double				eps)			// the error bound
{

	ANNkdDim = dim;						// copy arguments to static equivs
	ANNkdQ = q;
	ANNkdPts = pts;
#ifndef ANN_NO_WEIGHTS
	ANNkdWeights = weights;
#endif
#ifdef WANN_TORUS
    ANNkdL = L;
#endif
	ANNptsVisited = 0;					// initialize count of points visited

	if (k > n_pts) {					// too many near neighbors?
		annError("Requesting more near neighbors than data points", ANNabort);
	}

	ANNkdMaxErr = ANN_POW(1.0 + eps);
	ANN_FLOP(2)							// increment floating op count

	ANNkdPointMK = new ANNmin_k(k);		// create set for closest k points
										// search starting at the root
	root->ann_search(annBoxDistance(q, bnd_box_lo, bnd_box_hi, dim
#ifdef WANN_TORUS
    ,L
#endif
    )
#ifdef WANN_TORUS
       ,q
#endif    
    );

	for (int i = 0; i < k; i++) {		// extract the k-th closest points
		dd[i] = ANNkdPointMK->ith_smallest_key(i);
		nn_idx[i] = ANNkdPointMK->ith_smallest_info(i);
	}
	delete ANNkdPointMK;				// deallocate closest point set
}

#ifdef WANN_TORUS
//-------------------------------------------------------------------------
// torus search utilites - some functions to simplify code
//-------------------------------------------------------------------------
//simple replacement 
ANNdist upd_dist(ANNdist old_dist,
                 ANNdist old_diff,
                 ANNdist new_diff){
    return (ANNdist) ANN_SUM(old_dist,
				ANN_DIFF(ANN_POW(old_diff), ANN_POW(new_diff)));    
}        

ANNbool not_need_inversion(ANNcoord point,
                        ANNcoord btm,
                        ANNcoord top,
                        ANNidx cut_dim){
    //will the inversion decrease the distance to the node?
    if (point < btm) {
		if (point+ANNkdL[cut_dim] - top>btm - point)
			return ANNtrue;
		else return ANNfalse;
	} else if (point >top) {
		if (btm - point+ ANNkdL[cut_dim]> point - top)
			return ANNtrue;
		else return ANNfalse;
	} else return ANNtrue; //point in the box - no inversion    
}
#endif

//----------------------------------------------------------------------
//	kd_split::ann_search - search a splitting node
//----------------------------------------------------------------------

void ANNkd_split::ann_search(ANNdist box_dist
#ifdef WANN_TORUS
       ,ANNpoint			q
#endif
)
{
										// check dist calc term condition
	if (ANNmaxPtsVisited != 0 && ANNptsVisited > ANNmaxPtsVisited) return;

#ifdef WANN_TORUS   
// in toroidal case ANNkdQ holds the true query point, 
// while particular nodes may find it more fruitful to work
// with their own custom point inverted along one or more axis

    ANNcoord cut_diff = q[cut_dim] - cut_val;
    ANNidx first_child;		// first child node to visit
    ANNidx second_child;	// second child node to visit
    ANNdist first_dist;		// distance to the first node
    ANNdist second_dist;	// distance to the node child[second_child]
    ANNpoint first_point;	// point to be passed to the first node
    ANNpoint second_point;	// for the second one

    //points that will be passed to LO and HI nodes 
    //(may be either first or second!)
    ANNpoint lo_point = annCopyPt(ANNkdDim, q);
    ANNpoint hi_point = annCopyPt(ANNkdDim, q);
    ANNdist lo_dist = box_dist;
    ANNdist hi_dist = box_dist; 

    //three choices that determine where to go

    //first choice: do we need to invert lo_point
    if (not_need_inversion(q[cut_dim], cd_bnds[ANN_LO], cut_val, cut_dim)) {
        //lo_point: no changes
        if (cut_diff > 0) { //only in this case we need to update distance
    		ANNcoord box_diff = lo_point[cut_dim] - cd_bnds[ANN_HI];
	    	if (box_diff < 0)				// within bounds - ignore
		    	box_diff = 0;
		    lo_dist=upd_dist(box_dist, box_diff, cut_diff);
        }		
    } else {
        //invert lo_point
        if (cut_diff < 0) { 
            lo_point[cut_dim] = lo_point[cut_dim] + ANNkdL[cut_dim];
     		ANNcoord box_diff = cd_bnds[ANN_LO] - q[cut_dim];
	    	if (box_diff < 0) {				// within bounds - ignore
		    	box_diff = 0; // cannot be in lo box		    	
		    	cerr << "lo_point was in lo_box before inversion!!!" << endl;
             }	
		    lo_dist = upd_dist(box_dist, box_diff, lo_point[cut_dim] - cut_val);
        } else {
            lo_point[cut_dim] = lo_point[cut_dim] - ANNkdL[cut_dim];
     		ANNcoord box_diff = q[cut_dim] - cd_bnds[ANN_HI];
	    	if (box_diff < 0)				// within bounds - ignore
		    	box_diff = 0;//here we may be in the upper box - if it is large enough
		    lo_dist = upd_dist(box_dist, box_diff, cd_bnds[ANN_LO] - lo_point[cut_dim]);	
        }    
    }          
    //second choice: do we need to invert hi_point?
    //much like the prevoius choice excepot for replaces like hi<-lo and dependent
        if (not_need_inversion(q[cut_dim], cut_val, cd_bnds[ANN_HI], cut_dim)) {
        //hi_point: no changes
        if (cut_diff < 0) { //only in this case we need to update distance
    		ANNcoord box_diff = cd_bnds[ANN_LO] - hi_point[cut_dim];
	    	if (box_diff < 0)				// within bounds - ignore
		    	box_diff = 0;
		    hi_dist = upd_dist(box_dist, box_diff, cut_diff);
        }		
    } else {
        //invert hi_point
        if (cut_diff > 0) { 
            hi_point[cut_dim] = hi_point[cut_dim] - ANNkdL[cut_dim];
     		ANNcoord box_diff = q[cut_dim] - cd_bnds[ANN_HI];
	    	if (box_diff < 0) {				// within bounds - ignore
		    	box_diff = 0; // cannot be in hi box because inner points are not inverted
		    	cerr << "hi_point was in hi_box before inversion!!!" << endl;
             }	
		    hi_dist = upd_dist(box_dist, box_diff, cut_val - hi_point[cut_dim]);
        } else {
            hi_point[cut_dim] = hi_point[cut_dim] + ANNkdL[cut_dim];
     		ANNcoord box_diff = cd_bnds[ANN_LO] - q[cut_dim];
	    	if (box_diff < 0)				// within bounds - ignore
		    	box_diff = 0; //here we may be in the upper box - if it is large enough
		    hi_dist = upd_dist(box_dist, box_diff, hi_point[cut_dim] - cd_bnds[ANN_HI]);
        }    
    }          
    //third choice: where do you want to go today?
#ifndef ANN_NO_WEIGHTS
	if (lo_dist + child[ANN_LO]->minWeight <
		hi_dist + child[ANN_HI]->minWeight)
	{								// low child first
#else
    if (lo_dist<hi_dist) { // condition compatible with ANN_WEIGHTS 
#endif    
        first_child   = ANN_LO;
        first_point   = lo_point;
        first_dist    = lo_dist;
        second_child = ANN_HI;
        second_point = hi_point;
        second_dist  = hi_dist;        
    } else {
        first_child   = ANN_HI;
        first_point   = hi_point;
        first_dist    = hi_dist;
        second_child = ANN_LO;
        second_point = lo_point;
        second_dist  = lo_dist;                  
    }
    child[first_child]->ann_search(first_dist, first_point);   
    //check that we should visit another child
#ifndef ANN_NO_WEIGHTS
    if ((second_dist + child[second_child]->minWeight)*ANNkdMaxErr < ANNkdPointMK->max_key()) {
#else    
    if (second_dist*ANNkdMaxErr < ANNkdPointMK->max_key()) {
#endif                    
        child[second_child]->ann_search(second_dist, second_point);
}
    first_point = NULL;
    second_point =NULL;
    delete[] lo_point;
    delete[] hi_point;
#else // !WANN_TORUS: a flat open space
	// distance to cutting plane
	ANNcoord cut_diff = ANNkdQ[cut_dim] - cut_val;

#ifndef ANN_NO_WEIGHTS
	ANNdist lo_box_dist, hi_box_dist;
	if (cut_diff < 0) {					// left of cutting plane
		lo_box_dist = box_dist;			// low box closer
										// find dist to farther box
		ANNcoord box_diff = cd_bnds[ANN_LO] - ANNkdQ[cut_dim];
		if (box_diff < 0)				// within bounds - ignore
			box_diff = 0;
		hi_box_dist = (ANNdist) ANN_SUM(box_dist,
				ANN_DIFF(ANN_POW(box_diff), ANN_POW(cut_diff)));
	}
	else {								// right of cutting plane
		hi_box_dist = box_dist;			// high box closer
										// find dist to farther box
		ANNcoord box_diff = ANNkdQ[cut_dim] - cd_bnds[ANN_HI];
		if (box_diff < 0)				// within bounds - ignore
			box_diff = 0;
		lo_box_dist = (ANNdist) ANN_SUM(box_dist,
				ANN_DIFF(ANN_POW(box_diff), ANN_POW(cut_diff)));
	}
									// choose which child to visit first
	if (lo_box_dist + child[ANN_LO]->minWeight <
		hi_box_dist + child[ANN_HI]->minWeight)
	{								// low child first
		child[ANN_LO]->ann_search(lo_box_dist);
									// shall we visit the other one?
		if ((hi_box_dist + child[ANN_HI]->minWeight) * ANNkdMaxErr <
			ANNkdPointMK->max_key())
			child[ANN_HI]->ann_search(hi_box_dist);
	}
	else {							// high child first
		child[ANN_HI]->ann_search(hi_box_dist);
									// shall we visit the other one?
		if ((lo_box_dist + child[ANN_LO]->minWeight) * ANNkdMaxErr <
			ANNkdPointMK->max_key())
			child[ANN_LO]->ann_search(lo_box_dist);
	}
	ANN_FLOP(12)
#else // ANN_NO_WEIGHTS
	if (cut_diff < 0) {					// left of cutting plane
		child[ANN_LO]->ann_search(box_dist);// visit closer child first

		ANNcoord box_diff = cd_bnds[ANN_LO] - ANNkdQ[cut_dim];
		if (box_diff < 0)				// within bounds - ignore
			box_diff = 0;
										// distance to further box
		box_dist = (ANNdist) ANN_SUM(box_dist,
				ANN_DIFF(ANN_POW(box_diff), ANN_POW(cut_diff)));

										// visit further child if close enough
		if (box_dist * ANNkdMaxErr < ANNkdPointMK->max_key())
			child[ANN_HI]->ann_search(box_dist);

	}
	else {								// right of cutting plane
		child[ANN_HI]->ann_search(box_dist);// visit closer child first

		ANNcoord box_diff = ANNkdQ[cut_dim] - cd_bnds[ANN_HI];
		if (box_diff < 0)				// within bounds - ignore
			box_diff = 0;
										// distance to further box
		box_dist = (ANNdist) ANN_SUM(box_dist,
				ANN_DIFF(ANN_POW(box_diff), ANN_POW(cut_diff)));

										// visit further child if close enough
		if (box_dist * ANNkdMaxErr < ANNkdPointMK->max_key())
			child[ANN_LO]->ann_search(box_dist);

	}
	ANN_FLOP(10)						// increment floating ops
#endif // ANN_NO_WEIGHTS
	ANN_SPL(1)							// one more splitting node visited
#endif // !WANN_TORUS
}

//----------------------------------------------------------------------
//	kd_leaf::ann_search - search points in a leaf node
//		Note: The unreadability of this code is the result of
//		some fine tuning to replace indexing by pointer operations.
//----------------------------------------------------------------------

void ANNkd_leaf::ann_search(ANNdist box_dist
#ifdef WANN_TORUS
       ,ANNpoint			q
#endif
)
{
	register ANNdist dist;				// distance to data point
	register ANNcoord* pp;				// data coordinate pointer
	register ANNcoord* qq;				// query coordinate pointer
	register ANNdist min_dist;			// distance to k-th closest point
	register ANNcoord t;
	register int d;

	min_dist = ANNkdPointMK->max_key(); // k-th smallest distance so far

	for (int i = 0; i < n_pts; i++) {	// check points in bucket

		pp = ANNkdPts[bkt[i]];			// first coord of next data point
		qq = ANNkdQ;					// first coord of query point

// WANN_TORUS: here we use true points instead of (probably) inverted
// ones the reason is that given a large enough bucket size, one can
// encounter both points that require inversion and not

#ifndef ANN_NO_WEIGHTS
		dist = ANNkdWeights[bkt[i]];
#else
		dist = 0;
#endif

		for(d = 0; d < ANNkdDim; d++) {
			ANN_COORD(1)				// one more coordinate hit
			ANN_FLOP(4)					// increment floating ops

			t = *(qq++) - *(pp++);		// compute length and adv coordinate
#ifdef WANN_TORUS	
            if(t>ANNkdL[d]/2)               // the true distance in toroidal space
                t = t - ANNkdL[d];
            else if(t<-ANNkdL[d]/2)    
                t = t+ANNkdL[d];
#endif		
										// exceeds dist to k-th smallest?
			if( (dist = ANN_SUM(dist, ANN_POW(t))) > min_dist) {
				break;
			}
		}

		if (d >= ANNkdDim &&					// among the k best?
		   (ANN_ALLOW_SELF_MATCH || dist!=0)) { // and no self-match problem
												// add it to the list
			ANNkdPointMK->insert(dist, bkt[i]);
			min_dist = ANNkdPointMK->max_key();
		}
	}
	ANN_LEAF(1)							// one more leaf node visited
	ANN_PTS(n_pts)						// increment points visited
	ANNptsVisited += n_pts;				// increment number of points visited
}
