//----------------------------------------------------------------------
// File:			kd_tree.h
// Programmer:		Sunil Arya and David Mount,
//					Andrei Sobolevskii and Aleksandr Andrievskii
// Description:		Declarations for standard kd-tree routines
// Last modified:	06/22/08
//----------------------------------------------------------------------
// Copyright 2008 Andrei Sobolevskii and Aleksandr Andrievskii
//
// Copyright (c) 1997-2005 University of Maryland and Sunil Arya and
// David Mount.  All Rights Reserved.
// 
// This file is part of the WANN (Weighted Approximate Nearest 
// Neighbors) library.
//
// WANN is free software: you can redistribute it and/or modify it
// under the terms of the GNU Lesser General Public License as
// published by the Free Software Foundation, either version 3 of the
// License, or (at your option) any later version.
//
// WANN is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
// or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General
// Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public
// License along with WANN.  If not, see <http://www.gnu.org/licenses/>.
//----------------------------------------------------------------------
// Acknowledgment:
//
// The development of the WANN library has been supported by ANR
// (Agence Nationale de la Recherche, France) under contract 
// BLAN07-2_183172 OTARIE and by RFBR (Russian Foundation for Basic 
// Research) under contract RFBR/CNRS 07-01-92217.
//----------------------------------------------------------------------
// History:
//
// This file has been originally released under GNU LGPL as part of the ANN
// (Approximate Nearest Neighbors) library of David Mount and Sunil
// Arya.  It has been modified to become part of the WANN library.
//
// WANN history:
//	Revision 0.2 06/22/08
//		New arg q to ann_search, ann_pri_search, ann_FR_search
//	Revision 0.1 06/03/08
//		Modified copyright information.
//		Changes comments.
//		Added public members minWeight to ANNkd_node.
//		Added public method set_weight to ANNkd_node and its descendants.
//		If ANN_NO_WEIGHTS is not defined move bkt from ANNkd_leaf 
//			to ANNkd_node.
//		Added arg pidx to ANNkd_split::ANNkd_split.
//		Added new arg wa to rkd_tree.
//		
// ANN history:
//	Revision 1.1  05/03/05
//		Added fixed radius kNN search
//	Revision 0.1  03/04/98
//		Initial release
//----------------------------------------------------------------------

#ifndef ANN_kd_tree_H
#define ANN_kd_tree_H

#include <wann/ANNx.h>					// all ANN includes

using namespace std;					// make std:: available

//----------------------------------------------------------------------
//	Generic kd-tree node
//
//		Nodes in kd-trees are of two types, splitting nodes which contain
//		splitting information (a splitting hyperplane orthogonal to one
//		of the coordinate axes) and leaf nodes which contain point
//		information (an array of points stored in a bucket).  This is
//		handled by making a generic class kd_node, which is essentially an
//		empty shell, and then deriving the leaf and splitting nodes from
//		this.
//----------------------------------------------------------------------

class ANNkd_node{						// generic kd-tree node (empty shell)
public:
#ifndef ANN_NO_WEIGHTS
	ANNdist minWeight;					// minimal weight of points in node
	ANNidxArray bkt;					// union of buckets in node
#endif
	virtual ~ANNkd_node() {}					// virtual distroyer

	virtual void ann_search(ANNdist box_dist	// tree search
#ifdef WANN_TORUS
							, ANNpoint q = NULL
#endif
    ) = 0;		
	virtual void ann_pri_search(ANNdist box_dist// priority search
#ifdef WANN_TORUS
								, ANNpoint q = NULL
#endif
    ) = 0;	
	virtual void ann_FR_search(ANNdist box_dist	// fixed-radius search
#ifdef WANN_TORUS
							   , ANNpoint q = NULL
#endif
    ) = 0;	

	virtual void getStats(						// get tree statistics
				int dim,						// dimension of space
				ANNkdStats &st,					// statistics
				ANNorthRect &bnd_box) = 0;		// bounding box
												// print node
	virtual void print(int level, ostream &out) = 0;
	virtual void dump(ostream &out) = 0;		// dump node
#ifndef ANN_NO_WEIGHTS
	virtual void set_weight					// set weight of point
		(ANNidx* idx_ptr, ANNdist new_weight, ANNdistArray weights) = 0;
#endif

	friend class ANNkd_tree;					// allow kd-tree to access us
};

//----------------------------------------------------------------------
//	kd-splitting function:
//		kd_splitter is a pointer to a splitting routine for preprocessing.
//		Different splitting procedures result in different strategies
//		for building the tree.
//----------------------------------------------------------------------

typedef void (*ANNkd_splitter)(			// splitting routine for kd-trees
	ANNpointArray		pa,				// point array (unaltered)
	ANNidxArray			pidx,			// point indices (permuted on return)
	const ANNorthRect	&bnds,			// bounding rectangle for cell
	int					n,				// number of points
	int					dim,			// dimension of space
	int					&cut_dim,		// cutting dimension (returned)
	ANNcoord			&cut_val,		// cutting value (returned)
	int					&n_lo);			// num of points on low side (returned)

//----------------------------------------------------------------------
//	Leaf kd-tree node
//		Leaf nodes of the kd-tree store the set of points associated
//		with this bucket, stored as an array of point indices.  These
//		are indices in the array points, which resides with the
//		root of the kd-tree.  We also store the number of points
//		that reside in this bucket.
//----------------------------------------------------------------------

class ANNkd_leaf: public ANNkd_node		// leaf node for kd-tree
{
	int					n_pts;			// no. points in bucket
#ifdef ANN_NO_WEIGHTS
	ANNidxArray			bkt;			// bucket of points
#endif
public:
	ANNkd_leaf(							// constructor
		int				n,				// number of points
		ANNidxArray		b)				// bucket
		{
			n_pts		= n;			// number of points in bucket
			bkt			= b;			// the bucket
		}

	~ANNkd_leaf() { }					// destructor (none)

	virtual void getStats(						// get tree statistics
				int dim,						// dimension of space
				ANNkdStats &st,					// statistics
				ANNorthRect &bnd_box);			// bounding box
	virtual void print(int level, ostream &out);// print node
	virtual void dump(ostream &out);			// dump node

	virtual void ann_search(ANNdist box_dist	// standard search
#ifdef WANN_TORUS
							, ANNpoint			q = NULL
#endif
    );			
	virtual void ann_pri_search(ANNdist box_dist// priority search
#ifdef WANN_TORUS
								, ANNpoint			q = NULL
#endif
    );		
	virtual void ann_FR_search(ANNdist box_dist	// fixed-radius search
#ifdef WANN_TORUS
							   , ANNpoint			q = NULL
#endif
    );		
#ifndef ANN_NO_WEIGHTS
	virtual void set_weight					// update weight of point
		(ANNidx* idx_ptr, ANNdist new_weight, ANNdistArray weights);
#endif
};

//----------------------------------------------------------------------
//		KD_TRIVIAL is a special pointer to an empty leaf node. Since
//		some splitting rules generate many (more than 50%) trivial
//		leaves, we use this one shared node to save space.
//
//		The pointer is initialized to NULL, but whenever a kd-tree is
//		created, we allocate this node, if it has not already been
//		allocated. This node is *never* deallocated, so it produces
//		a small memory leak.
//----------------------------------------------------------------------

extern ANNkd_leaf *KD_TRIVIAL;					// trivial (empty) leaf node

//----------------------------------------------------------------------
//	kd-tree splitting node.
//		Splitting nodes contain a cutting dimension and a cutting value.
//		These indicate the axis-parellel plane which subdivide the
//		box for this node. The extent of the bounding box along the
//		cutting dimension is maintained (this is used to speed up point
//		to box distance calculations) [we do not store the entire bounding
//		box since this may be wasteful of space in high dimensions].
//		We also store pointers to the 2 children.
//----------------------------------------------------------------------

class ANNkd_split : public ANNkd_node	// splitting node of a kd-tree
{
	int					cut_dim;		// dim orthogonal to cutting plane
	ANNcoord			cut_val;		// location of cutting plane
	ANNcoord			cd_bnds[2];		// lower and upper bounds of
										// rectangle along cut_dim
	ANNkd_ptr			child[2];		// left and right children
public:
	ANNkd_split(						// constructor
		int cd,							// cutting dimension
		ANNcoord cv,					// cutting value
		ANNcoord lv, ANNcoord hv,				// low and high values
#ifndef ANN_NO_WEIGHTS
		ANNkd_ptr lc=NULL, ANNkd_ptr hc=NULL,	// children
		ANNidxArray pidx=NULL)					// union of buckets
#else
		ANNkd_ptr lc=NULL, ANNkd_ptr hc=NULL)	// children
#endif
		{
			cut_dim		= cd;					// cutting dimension
			cut_val		= cv;					// cutting value
			cd_bnds[ANN_LO] = lv;				// lower bound for rectangle
			cd_bnds[ANN_HI] = hv;				// upper bound for rectangle
			child[ANN_LO]	= lc;				// left child
			child[ANN_HI]	= hc;				// right child
#ifndef ANN_NO_WEIGHTS
			bkt = pidx;
#endif
		}

	~ANNkd_split()						// destructor
		{
			if (child[ANN_LO]!= NULL && child[ANN_LO]!= KD_TRIVIAL)
				delete child[ANN_LO];
			if (child[ANN_HI]!= NULL && child[ANN_HI]!= KD_TRIVIAL)
				delete child[ANN_HI];
		}

	virtual void getStats(						// get tree statistics
				int dim,						// dimension of space
				ANNkdStats &st,					// statistics
				ANNorthRect &bnd_box);			// bounding box
	virtual void print(int level, ostream &out);// print node
	virtual void dump(ostream &out);			// dump node

	virtual void ann_search(ANNdist box_dist	// standard search
#ifdef WANN_TORUS
							, ANNpoint		q = NULL
#endif
    );			
	virtual void ann_pri_search(ANNdist box_dist// priority search
#ifdef WANN_TORUS
								, ANNpoint		q = NULL
#endif
    );		
	virtual void ann_FR_search(ANNdist box_dist	// fixed-radius search
#ifdef WANN_TORUS
        , ANNpoint		q = NULL
#endif
    );		
#ifndef ANN_NO_WEIGHTS
	virtual void set_weight					// update weight of point
		(ANNidx* idx_ptr, ANNdist new_weight, ANNdistArray weights);
#endif
};

//----------------------------------------------------------------------
//		External entry points
//----------------------------------------------------------------------

ANNkd_ptr rkd_tree(				// recursive construction of kd-tree
	ANNpointArray		pa,				// point array (unaltered)
#ifndef ANN_NO_WEIGHTS
	ANNdistArray		wa,				// array of weights (unaltered)
#endif
	ANNidxArray			pidx,			// point indices to store in subtree
	int					n,				// number of points
	int					dim,			// dimension of space
	int					bsp,			// bucket space
	ANNorthRect			&bnd_box,		// bounding box for current node
	ANNkd_splitter		splitter);		// splitting routine

#endif
