//----------------------------------------------------------------------
// File:			kd_util.h
// Programmer:		Sunil Arya and David Mount, Aleksandr Andrievskii
// Description:		Common utilities for kd- trees
// Last modified:	06/22/08
//----------------------------------------------------------------------
// Copyright 2008 Aleksandr Andrievskii
//
// Copyright (c) 1997-2005 University of Maryland and Sunil Arya and
// David Mount.  All Rights Reserved.
// 
// This file is part of the WANN (Weighted Approximate Nearest 
// Neighbors) library.
//
// WANN is free software: you can redistribute it and/or modify it
// under the terms of the GNU Lesser General Public License as
// published by the Free Software Foundation, either version 3 of the
// License, or (at your option) any later version.
//
// WANN is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
// or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General
// Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public
// License along with WANN.  If not, see <http://www.gnu.org/licenses/>.
//----------------------------------------------------------------------
// Acknowledgment:
//
// The development of the WANN library has been supported by ANR
// (Agence Nationale de la Recherche, France) under contract 
// BLAN07-2_183172 OTARIE and by RFBR (Russian Foundation for Basic 
// Research) under contract RFBR/CNRS 07-01-92217.
//----------------------------------------------------------------------
// History:
//
// This file has been originally released under GNU LGPL as part of the ANN
// (Approximate Nearest Neighbors) library of David Mount and Sunil
// Arya.  It has been modified to become part of the WANN library.
//
// WANN history:
//	Revision 0.2 06/22/08
//		New arg L to annBoxDistance
//	Revision 0.1 06/03/08
//		Modified copying permission statement.
//		
// ANN history:
//	Revision 0.1  03/04/98
//		Initial release
//----------------------------------------------------------------------

#ifndef ANN_kd_util_H
#define ANN_kd_util_H

#include "kd_tree.h"					// kd-tree declarations

//----------------------------------------------------------------------
//	externally accessible functions
//----------------------------------------------------------------------

double annAspectRatio(			// compute aspect ratio of box
	int					dim,			// dimension
	const ANNorthRect	&bnd_box);		// bounding cube

void annEnclRect(				// compute smallest enclosing rectangle
	ANNpointArray		pa,				// point array
	ANNidxArray			pidx,			// point indices
	int					n,				// number of points
	int					dim,			// dimension
	ANNorthRect &bnds);					// bounding cube (returned)

void annEnclCube(				// compute smallest enclosing cube
	ANNpointArray		pa,				// point array
	ANNidxArray			pidx,			// point indices
	int					n,				// number of points
	int					dim,			// dimension
	ANNorthRect &bnds);					// bounding cube (returned)

ANNdist annBoxDistance(			// compute distance from point to box
	const ANNpoint		q,				// the point
	const ANNpoint		lo,				// low point of box
	const ANNpoint		hi,				// high point of box
	int					dim             // dimension of space
#ifdef WANN_TORUS
    ,const ANNdistArray  L = NULL       //size of space
#endif    
    );			

ANNcoord annSpread(				// compute point spread along dimension
	ANNpointArray		pa,				// point array
	ANNidxArray			pidx,			// point indices
	int					n,				// number of points
	int					d);				// dimension to check

void annMinMax(					// compute min and max coordinates along dim
	ANNpointArray		pa,				// point array
	ANNidxArray			pidx,			// point indices
	int					n,				// number of points
	int					d,				// dimension to check
	ANNcoord&			min,			// minimum value (returned)
	ANNcoord&			max);			// maximum value (returned)

int annMaxSpread(				// compute dimension of max spread
	ANNpointArray		pa,				// point array
	ANNidxArray			pidx,			// point indices
	int					n,				// number of points
	int					dim);			// dimension of space

void annMedianSplit(			// split points along median value
	ANNpointArray		pa,				// points to split
	ANNidxArray			pidx,			// point indices
	int					n,				// number of points
	int					d,				// dimension along which to split
	ANNcoord			&cv,			// cutting value
	int					n_lo);			// split into n_lo and n-n_lo

void annPlaneSplit(				// split points by a plane
	ANNpointArray		pa,				// points to split
	ANNidxArray			pidx,			// point indices
	int					n,				// number of points
	int					d,				// dimension along which to split
	ANNcoord			cv,				// cutting value
	int					&br1,			// first break (values < cv)
	int					&br2);			// second break (values == cv)

void annBoxSplit(				// split points by a box
	ANNpointArray		pa,				// points to split
	ANNidxArray			pidx,			// point indices
	int					n,				// number of points
	int					dim,			// dimension of space
	ANNorthRect			&box,			// the box
	int					&n_in);			// number of points inside (returned)

int annSplitBalance(			// determine balance factor of a split
	ANNpointArray		pa,				// points to split
	ANNidxArray			pidx,			// point indices
	int					n,				// number of points
	int					d,				// dimension along which to split
	ANNcoord			cv);			// cutting value

void annBox2Bnds(				// convert inner box to bounds
	const ANNorthRect	&inner_box,		// inner box
	const ANNorthRect	&bnd_box,		// enclosing box
	int					dim,			// dimension of space
	int					&n_bnds,		// number of bounds (returned)
	ANNorthHSArray		&bnds);			// bounds array (returned)

void annBnds2Box(				// convert bounds to inner box
	const ANNorthRect	&bnd_box,		// enclosing box
	int					dim,			// dimension of space
	int					n_bnds,			// number of bounds
	ANNorthHSArray		bnds,			// bounds array
	ANNorthRect			&inner_box);	// inner box (returned)

#endif
