//----------------------------------------------------------------------
//	File:			rand.h
//	Programmer:		Sunil Arya and David Mount
//	Description:	Basic include file for random point generators
//	Last modified:	06/03/08
//----------------------------------------------------------------------
// Copyright (c) 1997-2005 University of Maryland and Sunil Arya and
// David Mount.  All Rights Reserved.
// 
// This file is part of the WANN (Weighted Approximate Nearest 
// Neighbors) library.
//
// WANN is free software: you can redistribute it and/or modify it
// under the terms of the GNU Lesser General Public License as
// published by the Free Software Foundation, either version 3 of the
// License, or (at your option) any later version.
//
// WANN is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
// or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General
// Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public
// License along with WANN.  If not, see <http://www.gnu.org/licenses/>.
//----------------------------------------------------------------------
// Acknowledgment:
//
// The development of the WANN library has been supported by ANR
// (Agence Nationale de la Recherche, France) under contract 
// BLAN07-2_183172 OTARIE and by RFBR (Russian Foundation for Basic 
// Research) under contract RFBR/CNRS 07-01-92217.
//----------------------------------------------------------------------
// History:
//
// This file has been originally released under GNU LGPL as part of the ANN
// (Approximate Nearest Neighbors) library of David Mount and Sunil
// Arya.  It has been modified to become part of the WANN library.
//
// WANN history:
//	Revision 0.1 06/03/08
//		Modified copying permission.
//		Includes guarded by HAVE_CONFIG_H, HAVE_RANDOM, HAVE_RAND.
//
// ANN history:
//	Revision 0.1  03/04/98
//		Initial release
//	Revision 1.0  04/01/05
//		Added annClusOrthFlats distribution
//		Changed procedure names to avoid namespace conflicts
//		Added annClusFlats distribution
//	Revision 1.1.1  08/04/06
//		Added planted distribution
//----------------------------------------------------------------------

#ifndef rand_H
#define rand_H

//----------------------------------------------------------------------
//	Basic includes
//----------------------------------------------------------------------

#ifdef HAVE_CONFIG_H
#	include <config.h>
#	ifdef HAVE_STDLIB_H
#		include <cstdlib>				// standard includes (rand/random)
#	endif
#	ifdef HAVE_MATH_H
#		include <cmath>					// math includes
#	endif
#else
#	include <cstdlib>					// standard includes (rand/random)
#	include <cmath>						// math routines
#endif

#include <wann/ANN.h>					// basic ANN includes

//----------------------------------------------------------------------
//	Although random/srandom is a more reliable random number generator,
//	many systems do not have it.  If it is not available, substitute 
//	the use of rand/srand for them either setting either HAVE_RAND
//	automatically in configure.h (recommended) or by setting
//	ANN_NO_RANDOM manually (to maintain compatibility with the
//	previous version of the ANN library if the user decides to bypass
//	the configure script).  In case neither config.h exists nor
//	ANN_NO_RANDOM is defined, assume that random/srandom is available.
//----------------------------------------------------------------------

#ifndef HAVE_CONFIG_H
#	ifdef ANN_NO_RANDOM
#		define HAVE_RAND	1
#	else
#		define HAVE_RANDOM	1
#	endif
#endif

#ifdef HAVE_RANDOM						// for systems having random()
#	define ANN_RAND		random
#	define ANN_SRAND	srandom
#	define ANN_RAND_MAX	2147483647UL		// 2**{31} - 1
  // #define ANN_RAND_MAX	1073741824UL		// 2**{30}
#else
#	ifdef HAVE_RAND						// otherwise use rand()
#		define ANN_RAND		rand
#		define ANN_SRAND	srand
#		define ANN_RAND_MAX	RAND_MAX
#	else
#		error "test/rand.h unable to find random number generator"
#	endif
#endif

//----------------------------------------------------------------------
//	Globals
//----------------------------------------------------------------------
extern	int		annIdum;				// random number seed

//----------------------------------------------------------------------
//	External entry points
//----------------------------------------------------------------------

void annUniformPts(				// uniform distribution
	ANNpointArray	pa,			// point array (modified)
	int				n,			// number of points
	int				dim);		// dimension

void annGaussPts(				// Gaussian distribution
	ANNpointArray	pa,			// point array (modified)
	int				n,			// number of points
	int				dim,		// dimension
	double			std_dev);	// standard deviation

void annCoGaussPts(				// correlated-Gaussian distribution
	ANNpointArray	pa,			// point array (modified)
	int				n,			// number of points
	int				dim,		// dimension
	double			correlation);	// correlation

void annLaplacePts(				// Laplacian distribution
	ANNpointArray	pa,			// point array (modified)
	int				n,			// number of points
	int				dim);		// dimension

void annCoLaplacePts(			// correlated-Laplacian distribution
	ANNpointArray	pa,			// point array (modified)
	int				n,			// number of points
	int				dim,		// dimension
	double			correlation);	// correlation

void annClusGaussPts(			// clustered-Gaussian distribution
	ANNpointArray	pa,			// point array (modified)
	int				n,			// number of points
	int				dim,		// dimension
	int				n_clus,		// number of colors (clusters)
	ANNbool			new_clust,	// generate new cluster centers
	double			std_dev);	// standard deviation within clusters

void annClusOrthFlats(          // clustered along orthogonal flats
	ANNpointArray	pa,			// point array (modified)
	int				n,			// number of points
	int				dim,		// dimension
	int				n_clus,		// number of colors
	ANNbool			new_clust,	// generate new clusters.
	double			std_dev,	// standard deviation within clusters
	int				max_dim);	// maximum dimension of the flats

void annClusEllipsoids(			// clustered around ellipsoids
	ANNpointArray	pa,			// point array (modified)
	int				n,			// number of points
	int				dim,		// dimension
	int				n_clus,		// number of colors
	ANNbool			new_clust,	// generate new clusters.
	double			std_dev_small,	// small standard deviation
	double			std_dev_lo,	// low standard deviation for ellipses
	double			std_dev_hi,	// high standard deviation for ellipses
	int				max_dim);	// maximum dimension of the flats

void annPlanted(				// planted nearest neighbors
	ANNpointArray	pa,			// point array (modified)
	int				n,			// number of points
	int				dim,		// dimension
	ANNpointArray	src,		// source point array
	int				n_src,		// source size
	double			std_dev);	// standard deviation about source

#endif
